@@ -1,6 +1,5 @@
 from neomodel import (StructuredNode, RelationshipTo, RelationshipFrom,
         StringProperty, IntegerProperty)
-from neomodel.exception import NotConnected
 
 
 class Person(StructuredNode):
@@ -18,7 +17,7 @@ class Person(StructuredNode):
 
 class Country(StructuredNode):
     code = StringProperty(unique_index=True)
-    inhabitant = RelationshipFrom('Person', 'IS_FROM')
+    inhabitant = RelationshipFrom(Person, 'IS_FROM')
 
 
 class SuperHero(Person):
@@ -54,14 +53,6 @@ def test_bidirectional_relationships():
     assert s.name == 'Jim'
 
     u.is_from.disconnect(b)
-    try:
-        u.is_from.disconnect(b)
-    except NotConnected:
-        assert True
-    else:
-        assert False
-
-    assert not u.is_from.all()
     assert not u.is_from.is_connected(b)
 
 
@@ -85,23 +76,6 @@ def test_custom_methods():
     assert u.special_name == 'Joe91'
 
 
-def test_abstract_class_relationships():
-    j = Person(name='Joe', age=13).save()
-    assert j
-
-    u = SuperHero(name='UltraJoe', age=13, power='invisibility').save()
-    assert u
-
-    gr = Country(code='GR').save()
-    assert gr
-
-    gr.inhabitant.connect(j)
-    assert gr.inhabitant.is_connected(j)
-
-    gr.inhabitant.connect(u)
-    assert gr.inhabitant.is_connected(u)
-
-
 def test_props_relationship():
     u = Person(name='Mar', age=20).save()
     assert u
